/* AArch64 SME intrinsics include file.
   Copyright (C) 2023-2025 Free Software Foundation, Inc.

   This file is part of GCC.

   GCC is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 3, or (at your
   option) any later version.

   GCC is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   Under Section 7 of GPL version 3, you are granted additional
   permissions described in the GCC Runtime Library Exception, version
   3.1, as published by the Free Software Foundation.

   You should have received a copy of the GNU General Public License and
   a copy of the GCC Runtime Library Exception along with this program;
   see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
   <http://www.gnu.org/licenses/>.  */

#ifndef _ARM_SME_H_
#define _ARM_SME_H_

#include <arm_sve.h>
#pragma GCC aarch64 "arm_sme.h"

void __arm_za_disable(void) __arm_streaming_compatible;

void *__arm_sc_memcpy(void *, const void *, __SIZE_TYPE__)
  __arm_streaming_compatible;

void *__arm_sc_memmove(void *, const void *, __SIZE_TYPE__)
  __arm_streaming_compatible;

void *__arm_sc_memset(void *, int, __SIZE_TYPE__)
  __arm_streaming_compatible;

void *__arm_sc_memchr(void *, int, __SIZE_TYPE__)
  __arm_streaming_compatible;

#endif
