// std::placeholders -*- C++ -*-

// Copyright (C) 2011-2025 Free Software Foundation, Inc.
//
// This file is part of the GNU ISO C++ Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// Under Section 7 of GPL version 3, you are granted additional
// permissions described in the GCC Runtime Library Exception, version
// 3.1, as published by the Free Software Foundation.

// You should have received a copy of the GNU General Public License and
// a copy of the GCC Runtime Library Exception along with this program;
// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
// <http://www.gnu.org/licenses/>.

#if __cplusplus < 201103L
# error "placeholders.cc must be compiled with -std=gnu++0x"
#endif

#include <functional>

namespace std _GLIBCXX_VISIBILITY(default)
{
_GLIBCXX_BEGIN_NAMESPACE_VERSION

  namespace placeholders
  {
    extern const _Placeholder<1> _1{};
    extern const _Placeholder<2> _2{};
    extern const _Placeholder<3> _3{};
    extern const _Placeholder<4> _4{};
    extern const _Placeholder<5> _5{};
    extern const _Placeholder<6> _6{};
    extern const _Placeholder<7> _7{};
    extern const _Placeholder<8> _8{};
    extern const _Placeholder<9> _9{};
    extern const _Placeholder<10> _10{};
    extern const _Placeholder<11> _11{};
    extern const _Placeholder<12> _12{};
    extern const _Placeholder<13> _13{};
    extern const _Placeholder<14> _14{};
    extern const _Placeholder<15> _15{};
    extern const _Placeholder<16> _16{};
    extern const _Placeholder<17> _17{};
    extern const _Placeholder<18> _18{};
    extern const _Placeholder<19> _19{};
    extern const _Placeholder<20> _20{};
    extern const _Placeholder<21> _21{};
    extern const _Placeholder<22> _22{};
    extern const _Placeholder<23> _23{};
    extern const _Placeholder<24> _24{};
    extern const _Placeholder<25> _25{};
    extern const _Placeholder<26> _26{};
    extern const _Placeholder<27> _27{};
    extern const _Placeholder<28> _28{};
    extern const _Placeholder<29> _29{};
  }

_GLIBCXX_END_NAMESPACE_VERSION
}
